@@ -2,10 +2,10 @@
 
 __description__ = 'pdf-parser, use it to parse a PDF document'
 __author__ = 'Didier Stevens'
-__version__ = '0.7.8'
-__date__ = '2023/01/03'
+__version__ = '0.7.12'
+__date__ = '2025/04/21'
 __minimum_python_version__ = (2, 5, 1)
-__maximum_python_version__ = (3, 11, 1)
+__maximum_python_version__ = (3, 12, 2)
 
 """
 Source code put in public domain by Didier Stevens, no Copyright
@@ -74,6 +74,11 @@ History:
   2022/05/24: bug fixes
   2022/11/09: V0.7.7 added support for environment variable DSS_DEFAULT_HASH_ALGORITHMS
   2023/01/03: V0.7.8 added unreferenced objects to statistics
+  2024/03/21: V0.7.9 added option jsonoutput; added verbose YARA rules
+  2024/10/25: V0.7.10 /ObjStm fix (x9090 PR)
+  2024/10/26: added pyzipper support
+  2025/03/04: V0.7.11 regex string fix Python 3.12 xambroz
+  2025/04/21: V0.7.12 bugfix YARACompile
 
 Todo:
   - handle printf todo
@@ -87,10 +92,10 @@ import zlib
 import binascii
 import hashlib
 import sys
-import zipfile
 import time
 import os
 import textwrap
+import json
 if sys.version_info[0] >= 3:
     from io import StringIO
     import urllib.request
@@ -105,6 +110,10 @@ try:
     import yara
 except:
     pass
+try:
+    import pyzipper as zipfile
+except ImportError:
+    import zipfile
 
 CHAR_WHITESPACE = 1
 CHAR_DELIMITER = 2
@@ -145,6 +154,9 @@ Option -H calculates the MD5 hash by default.
 This can be changed by setting environment variable DSS_DEFAULT_HASH_ALGORITHMS.
 Like this: set DSS_DEFAULT_HASH_ALGORITHMS=sha256
 
+Option --jsonoutput produces JSON output with the stream content of all objects with streams. Options -f and --overridingfilters apply.
+For example, if option -f is used, the JSON output contains the filtered streams, otherwise the JSON output contains the unfiltered streams.
+
 '''
     for line in manual.split('\n'):
         print(textwrap.fill(line))
@@ -197,6 +209,12 @@ def CopyWithoutWhiteSpace(content):
 def Obj2Str(content):
     return ''.join(map(lambda x: repr(x[1])[1:-1], CopyWithoutWhiteSpace(content)))
 
+def CreateZipFileObject(arg1, arg2):
+    if 'AESZipFile' in dir(zipfile):
+        return zipfile.AESZipFile(arg1, arg2)
+    else:
+        return zipfile.ZipFile(arg1, arg2)
+
 class cPDFDocument:
     def __init__(self, file):
         self.file = file
@@ -214,7 +232,7 @@ class cPDFDocument:
                 sys.exit()
         elif file.lower().endswith('.zip'):
             try:
-                self.zipfile = zipfile.ZipFile(file, 'r')
+                self.zipfile = CreateZipFileObject(file, 'r')
                 self.infile = self.zipfile.open(self.zipfile.infolist()[0], 'r', C2BIP3('infected'))
             except:
                 print('Error opening file %s' % file)
@@ -1147,7 +1165,7 @@ def PrintGenerateObject(object, options, newId=None):
         if options.filter:
             decompressed = object.Stream(True, options.overridingfilters)
             if decompressed == 'No filters' or decompressed.startswith('Unsupported filter: '):
-                print('    oPDF.stream(%d, %d, %s, %s)' % (objectId, object.version, repr(object.Stream(False, options.overridingfilters).rstrip()), repr(re.sub('/Length\s+\d+', '/Length %d', FormatOutput(dataPrecedingStream, True)).strip())))
+                print('    oPDF.stream(%d, %d, %s, %s)' % (objectId, object.version, repr(object.Stream(False, options.overridingfilters).rstrip()), repr(re.sub(r'/Length\s+\d+', '/Length %d', FormatOutput(dataPrecedingStream, True)).strip())))
             else:
                 dictionary = FormatOutput(dataPrecedingStream, True)
                 dictionary = re.sub(r'/Length\s+\d+', '', dictionary)
@@ -1158,7 +1176,7 @@ def PrintGenerateObject(object, options, newId=None):
                 dictionary = dictionary.strip()
                 print("    oPDF.stream2(%d, %d, %s, %s, 'f')" % (objectId, object.version, repr(decompressed.rstrip()), repr(dictionary)))
         else:
-            print('    oPDF.stream(%d, %d, %s, %s)' % (objectId, object.version, repr(object.Stream(False, options.overridingfilters).rstrip()), repr(re.sub('/Length\s+\d+', '/Length %d', FormatOutput(dataPrecedingStream, True)).strip())))
+            print('    oPDF.stream(%d, %d, %s, %s)' % (objectId, object.version, repr(object.Stream(False, options.overridingfilters).rstrip()), repr(re.sub(r'/Length\s+\d+', '/Length %d', FormatOutput(dataPrecedingStream, True)).strip())))
     else:
         print('    oPDF.indirectobject(%d, %d, %s)' % (objectId, object.version, repr(FormatOutput(object.content, True).strip())))
 
@@ -1193,16 +1211,20 @@ def ProcessAt(argument):
 def YARACompile(ruledata):
     if ruledata.startswith('#'):
         if ruledata.startswith('#h#'):
-            rule = binascii.a2b_hex(ruledata[3:])
+            rule = binascii.a2b_hex(ruledata[3:]).decode('latin')
         elif ruledata.startswith('#b#'):
-            rule = binascii.a2b_base64(ruledata[3:])
+            rule = binascii.a2b_base64(ruledata[3:]).decode('latin')
         elif ruledata.startswith('#s#'):
             rule = 'rule string {strings: $a = "%s" ascii wide nocase condition: $a}' % ruledata[3:]
         elif ruledata.startswith('#q#'):
             rule = ruledata[3:].replace("'", '"')
+        elif ruledata.startswith('#x#'):
+            rule = 'rule hexadecimal {strings: $a = { %s } condition: $a}' % ruledata[3:]
+        elif ruledata.startswith('#r#'):
+            rule = 'rule regex {strings: $a = /%s/ ascii wide nocase condition: $a}' % ruledata[3:]
         else:
             rule = ruledata[1:]
-        return yara.compile(source=rule)
+        return yara.compile(source=rule), rule
     else:
         dFilepaths = {}
         if os.path.isdir(ruledata):
@@ -1213,7 +1235,7 @@ def YARACompile(ruledata):
         else:
             for filename in ProcessAt(ruledata):
                 dFilepaths[filename] = filename
-        return yara.compile(filepaths=dFilepaths)
+        return yara.compile(filepaths=dFilepaths), ','.join(dFilepaths.values())
 
 def AddDecoder(cClass):
     global decoders
@@ -1390,6 +1412,22 @@ def CalculateChosenHash(data):
     dHashes[hashes[0]].update(data)
     return dHashes[hashes[0]].hexdigest(), hashes[0]
 
+class cMyJSONOutput():
+
+    def __init__(self):
+        self.items = []
+        self.counter = 1
+
+    def AddIdItem(self, id, name, data):
+        self.items.append({'id': id, 'name': name, 'content': binascii.b2a_base64(data).strip(b'\n').decode()})
+
+    def AddItem(self, name, data):
+        self.AddIdItem(self.counter, name, data)
+        self.counter += 1
+
+    def GetJSON(self):
+        return json.dumps({'version': 2, 'id': 'didierstevens.com', 'type': 'content', 'fields': ['id', 'name', 'content'], 'items': self.items})
+
 def Main():
     """pdf-parser, use it to parse a PDF document
     """
@@ -1426,6 +1464,7 @@ def Main():
     oParser.add_option('--decoders', type=str, default='', help='decoders to load (separate decoders with a comma , ; @file supported)')
     oParser.add_option('--decoderoptions', type=str, default='', help='options for the decoder')
     oParser.add_option('-k', '--key', help='key to search in dictionaries')
+    oParser.add_option('-j', '--jsonoutput', action='store_true', default=False, help='produce json output')
     (options, args) = oParser.parse_args(GetArguments())
 
     if options.man:
@@ -1540,9 +1579,12 @@ def Main():
             if not 'yara' in sys.modules:
                 print('Error: option yara requires the YARA Python module.')
                 return
-            rules = YARACompile(options.yara)
+            rules, rulesVerbose = YARACompile(options.yara)
+            if options.verbose:
+                print(rulesVerbose)
 
         oPDFParserOBJSTM = None
+        oMyJSONOutput = cMyJSONOutput()
         while True:
             if oPDFParserOBJSTM == None:
                 object = oPDFParser.GetObject()
@@ -1556,7 +1598,7 @@ def Main():
                 oPDFParseDictionary = cPDFParseDictionary(object.ContainsStream(), options.nocanonicalizedoutput)
                 numberOfObjects = int(oPDFParseDictionary.Get('/N')[0])
                 offsetFirstObject = int(oPDFParseDictionary.Get('/First')[0])
-                indexes = list(map(int, C2SIP3(object.Stream())[:offsetFirstObject].strip().split(' ')))
+                indexes = list(map(int, C2SIP3(object.Stream())[:offsetFirstObject].strip().replace('\n', ' ').split(' ')))
                 if len(indexes) % 2 != 0 or len(indexes) / 2 != numberOfObjects:
                     raise Exception('Error in index of /ObjStm stream')
                 streamObject = C2SIP3(object.Stream()[offsetFirstObject:])
@@ -1600,11 +1642,18 @@ def Main():
                         for reference in object.GetReferences():
                             objectsReferenced.add(reference)
                         objectsAll.add((str(object.id), str(object.version), 'R'))
+                elif options.jsonoutput:
+                    if object.type == PDF_ELEMENT_INDIRECT_OBJECT:
+                        if object.ContainsStream():
+                            filtered = object.Stream(options.filter == True, options.overridingfilters)
+                            if filtered == []:
+                                filtered = ''
+                            oMyJSONOutput.AddItem('obj %s %s' % (object.id, object.version), C2BIP3(filtered))
                 else:
                     if object.type == PDF_ELEMENT_COMMENT and selectComment:
                         if options.generate:
                             comment = object.comment[1:].rstrip()
-                            if re.match('PDF-\d\.\d', comment):
+                            if re.match(r'PDF-\d\.\d', comment):
                                 print("    oPDF.header('%s')" % comment[4:])
                             elif comment != '%EOF':
                                 print('    oPDF.comment(%s)' % repr(comment))
@@ -1731,6 +1780,9 @@ def Main():
                     if len(dKeywords[keyword]) > 0:
                         print(' %s %d: %s' % (keyword, len(dKeywords[keyword]), ', '.join(map(lambda x: '%d' % x, dKeywords[keyword]))))
 
+        if options.jsonoutput:
+            print(oMyJSONOutput.GetJSON())
+
         if options.generate or options.generateembedded != 0:
             print("    oPDF.xrefAndTrailer('%s')" % ' '.join(savedRoot))
             print('')
